// ---------------------------------------------------------------------------
// - Input.cpp                                                               -
// - standard object library - input stream class implementation             -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2008 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Byte.hpp"
#include "Input.hpp"
#include "Ascii.hpp"
#include "Vector.hpp"
#include "Unicode.hpp"
#include "Boolean.hpp"
#include "Integer.hpp"
#include "Character.hpp"
#include "QuarkZone.hpp"
#include "Exception.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - private section                                                       -
  // -------------------------------------------------------------------------

  // the read blocking mode
  static const long DEF_BLOK_MODE = -1;

  // this procedure encode a string based on a mode
  static char* stoc (const Encoding::t_emod emod, const String& s) {
    t_quad* sbuf = s.toquad ();
    try {
      char* cbuf = Unicode::encode (emod, sbuf);
      delete [] sbuf;
      return cbuf;
    } catch (...) {
      delete [] sbuf;
      throw;
    }
  }

  // -------------------------------------------------------------------------
  // - class section                                                         -
  // -------------------------------------------------------------------------

  // create a default input object

  Input::Input (void) {
    d_tout = DEF_BLOK_MODE;
    d_sbuf.reset ();
  }

  // flush the input stream buffer

  void Input::flush (void) {
    wrlock ();
    try {
      d_sbuf.reset ();
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // flush the input stream until a character is reached

  void Input::flush (const char value) {
    wrlock ();
    try {
      while (valid () == true) {
	char c = read ();
	if (c == value) break;
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // flush the input stream until a character is reached

  void Input::flush (const t_quad value) {
    wrlock ();
    try {
      while (valid () == true) {
	  t_quad c = getu ();
	if (c == value) break;
      }
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // return the stream descriptor

  int Input::getsid (void) const {
    return -1;
  }

  // get the stream timeout

  long Input::gettout (void) const {
    rdlock ();
    long result = d_tout;
    unlock ();
    return result;
  }

  // set the stream timeout

  void Input::settout (const long tout) {
    wrlock ();
    d_tout = tout;
    unlock ();
  }

  // read the next unicode character

  t_quad Input::getu (void) {
    wrlock ();
    try {
      // process byte mode
      if (d_emod == Encoding::BYTE) {
	// check for timeout
	if (valid () == false) {
	  if (iseos () == true) {
	    t_quad result = encode (eosq);
	    unlock ();
	    return result;
	  }
	  throw Exception ("read-error", "read byte character timeout");
	}
	// read and encode character
	t_quad result = encode (read ());
	unlock ();
	return result;
      }

      // process utf mode
      if (d_emod == Encoding::UTF8) {
	// the utf byte buffer
	char ubuf[Unicode::MAX_UTF8_SIZE];
	// the read status
	bool status = true;
	// loop until full
	for (long i = 0; i < Unicode::MAX_UTF8_SIZE; i++) {
	  // check for timeout
	  if (valid () == false) {
	    if ((iseos () == true) && (status == true)) {
	      unlock ();
	      return eosq;
	    }
	    throw Exception ("read-error", "read unicode character timeout");
	  }
	  // read the character
	  ubuf[i] = read ();
	  // check for validity
	  status = Unicode::valid (d_emod, ubuf, i+1);
	  if (status == true) {
	    t_quad result = Unicode::decode (ubuf);
	    unlock ();
	    return result;
	  }
	}
	// invalid character read
	throw Exception ("read-error", "cannot read unicode character");
      }
      // internal error
      throw Exception ("internal-error", "unicode character read mode");
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // read a buffer by size

  Buffer* Input::read (const long size) {
    wrlock ();
    Buffer* result = new Buffer;
    try {
      for (long i = 0; i < size; i++) {
	if (valid () == false) break;
	result->add (read ());
      }
    } catch (...) {
      delete result;
      unlock ();
      throw;
    }
    unlock ();
    return result;
  }

  // read a line from this input stream

  String Input::readln (void) {
    wrlock ();
    try {
      // create a buffer to accumulate characters
      Buffer buf;
      bool   flg = false;
      // read the character in the buffer
      while (valid () == true) {
	t_quad c = getu ();
	if (c == crlq) {
	  flg = true;
	  continue;
	}
	if (c == eolq) break;
	if (flg == true) {
	  buf.add (crlq);
	  flg = false;
	}
	buf.add (c);
      }
      unlock ();
      return buf.tostring ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // pushback a character

  void Input::pushback (const char value) {
    wrlock ();
    try {
      d_sbuf.pushback (value);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // pushback a c-string on this stream

  void Input::pushback (const char* s) {
    wrlock ();
    try {
      long size = Ascii::strlen (s);
      d_sbuf.pushback (s, size);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // pushback a c-string on this stream

  void Input::pushback (const char* s, const long size) {
    wrlock ();
    try {
      d_sbuf.pushback (s, size);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }
  
  // pushback a string on this stream
  
  void Input::pushback (const String& s) {
    wrlock ();
    try {
      // check for byte mode
      if (d_emod == Encoding::BYTE) {
	// get the quad buffer
	t_quad* sbuf = s.toquad ();
	if (sbuf == nilp) {
	  unlock ();
	  return;
	}
	// transcode to the character buffer
	char* cbuf = decode (sbuf);
	delete [] sbuf;
	// write the buffer and clean
	pushback (cbuf);
	delete [] cbuf;
	unlock ();
	return;
      }
      // check for utf-8 mode
      if (d_emod == Encoding::UTF8) {
	char* cbuf = stoc (d_emod, s);
	if (cbuf == nilp) {
	  unlock ();
	  return;
	}
	// write string content
	pushback (cbuf);
	delete [] cbuf;
	unlock ();
	return;
      }
      throw Exception ("pushback-error", "invalid mode for string pushback");
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // return the size of the input buffer

  long Input::buflen (void) const {
    rdlock ();
    try {
      long result = d_sbuf.length ();
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // return a string representation of the stream buffer

  String Input::tostring (void) const {
    rdlock ();
    try {
      String result = d_sbuf.tostring ();
      unlock ();
      return result;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // -------------------------------------------------------------------------
  // - object section                                                        -
  // -------------------------------------------------------------------------

  // the quark zone
  static const long QUARK_ZONE_LENGTH = 11;
  static QuarkZone  zone (QUARK_ZONE_LENGTH);

  // the object supported quarks
  static const long QUARK_EOSP     = zone.intern ("eos-p");
  static const long QUARK_READ     = zone.intern ("read");
  static const long QUARK_GETU     = zone.intern ("getu");
  static const long QUARK_FLUSH    = zone.intern ("flush");
  static const long QUARK_PUSHB    = zone.intern ("pushback");
  static const long QUARK_VALIDP   = zone.intern ("valid-p");
  static const long QUARK_READLN   = zone.intern ("readln");
  static const long QUARK_GETOUT   = zone.intern ("get-timeout");
  static const long QUARK_SETOUT   = zone.intern ("set-timeout");
  static const long QUARK_BUFLEN   = zone.intern ("get-buffer-length");
  static const long QUARK_TOSTRING = zone.intern ("to-string");

  // return true if the given quark is defined

  bool Input::isquark (const long quark, const bool hflg) const {
    rdlock ();
    if (zone.exists (quark) == true) {
      unlock ();
      return true;
    }
    bool result = hflg ? Stream::isquark (quark, hflg) : false;
    unlock ();
    return result;
  }

  // apply this object with a set of arguments and a quark

  Object* Input::apply (Runnable* robj, Nameset* nset, const long quark,
			Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_READ)     return new Byte      (read     ());
      if (quark == QUARK_GETU)     return new Character (getu     ());
      if (quark == QUARK_EOSP)     return new Boolean   (iseos    ());
      if (quark == QUARK_VALIDP)   return new Boolean   (valid    ());
      if (quark == QUARK_READLN)   return new String    (readln   ());
      if (quark == QUARK_GETOUT)   return new Integer   (gettout  ());
      if (quark == QUARK_BUFLEN)   return new Integer   (buflen   ());
      if (quark == QUARK_TOSTRING) return new String    (tostring ());
      if (quark == QUARK_FLUSH) {
	flush ();
	return nilp;
      }
    }

    // dispatch 1 argument
    if (argc == 1) {
     if (quark == QUARK_FLUSH) {
	t_quad c = argv->getchar (0);
	flush (c);
	return nilp;
      }
      if (quark == QUARK_SETOUT) {
	long tout = argv->getint (0);
	settout (tout);
	return nilp;
      }

      if (quark == QUARK_READ) {
	long size = argv->getint (0);
	return read (size);
      }
      if (quark == QUARK_PUSHB) {
	Object* obj = argv->get (0);
	// check for a byte
	Byte* bobj = dynamic_cast <Byte*> (obj);
	if (bobj != nilp) {
	  pushback (bobj->tobyte ());
	  return nilp;
	}
	// check for a character
	Character* cobj = dynamic_cast <Character*> (obj);
	if (cobj != nilp) {
	  pushback (cobj->tochar ());
	  return nilp;
	}
	String* s = dynamic_cast <String*> (obj);
	if (s != nilp) {
	  pushback (*s);
	  return nilp;
	}
	throw Exception ("type-error", "invalid object with pushback method",
			 Object::repr (obj));
      }
    }
    // apply these arguments with the stream object
    return Stream::apply (robj, nset, quark, argv);
  }
}
